@@ -2,9 +2,11 @@
 import sys
 import os
 import json
+import pickle
 import multiprocessing
 import random
 import subprocess
+import errno
 try:
     import ConfigParser as configparser
 except ImportError:
@@ -14,54 +16,51 @@ from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 
 
+
 import pytrailer as amt
 from .qtcustom import *
 from .qtcustom import resources
 from .downloader import TrailerDownloader, DownloadStatus
 
+
 class PyTrailerWidget(QMainWindow):
+    """Main PyQTrailer window. This is where it all starts"""
+
     configPath = '%s/.pyqtrailer' % os.path.expanduser('~')
+    cachePath = '%s/.pyqtrailer.cache' % os.path.expanduser('~')
 
     def __init__(self, *args):
         QMainWindow.__init__(self, *args)
+        # these are all categories apple is providing for now
         self.categories = [(self.tr('Just added'), '/trailers/home/feeds/just_added.json'),
                      (self.tr('Exclusive'), '/trailers/home/feeds/exclusive.json'),
                      (self.tr('Only HD'), '/trailers/home/feeds/just_hd.json'),
                      (self.tr('Most popular'), '/trailers/home/feeds/most_pop.json'),
                      (self.tr('Search'), '/trailers/home/scripts/quickfind.php?&q=')]
+
+        # pick sane defaults
         self.config = configparser.SafeConfigParser({'downloadDir':'/tmp',
                                        'filters':json.dumps([y for x, y in PyTrailerSettings.filters]),
                                        'readAhead':'4',
                                        'parallelDownload':'2',
                                        'player':'mplayer -user-agent %%a %%u'})
-        readAhead = int(self.config.get("DEFAULT","readAhead"))
-        self.config.read(self.configPath)
+
+        # run initializations
+        self.list_loader = None
+        self.list_loader_p = None
         self.movieDict = {}
-        self.readAheadTaskQueue = multiprocessing.Queue()
-        self.readAheadDoneQueue = multiprocessing.Queue()
-        self.trailerDownloadQueue = multiprocessing.Queue()
-        self.trailerDownloadDict = multiprocessing.Manager().dict()
-
-        self.readAheadProcess = []
-        for i in range(readAhead):
-            p = multiprocessing.Process(target=movieReadAhead,
-                        args=(self.readAheadTaskQueue,
-                              self.readAheadDoneQueue))
-            p.start()
-            self.readAheadProcess.append(p)
-
-        self.downloader = TrailerDownloader(self.trailerDownloadQueue,
-                               self.trailerDownloadDict,
-                               int(self.config.get("DEFAULT","parallelDownload")))
-        self.downloader.start()
+        self.config.read(self.configPath)
+        self.load_cache()
+        self.init_preloaders()
         self.init_widget()
         self.init_menus()
+        self.downloader.start()
 
     def init_widget(self):
+        """Initialize main child widgets, layouts etc."""
 
         self.refreshTimer = QTimer(self)
-        self.list_loader = None
-        self.refreshTimer.timeout.connect(self.refreshMovies)
+        self.refreshTimer.timeout.connect(self.refresh_movies)
         self.refreshTimer.start(1000)
         self.setWindowTitle(self.tr("PyTrailer - Apple Trailer Downloader"))
 
@@ -75,7 +74,7 @@ class PyTrailerWidget(QMainWindow):
             group.addButton(but)
             hbox.addWidget(but)
 
-        group.buttonClicked.connect(self.groupChange)
+        group.buttonClicked.connect(self.group_change)
 
         self.scrollArea = QScrollArea(centralWidget)
         scrollArea = self.scrollArea
@@ -116,7 +115,7 @@ class PyTrailerWidget(QMainWindow):
         scrollArea.setMinimumSize(QSize(400,350))
 
         self.mainArea = mlistLayout
-        self.loadGroup(self.tr("Just added"))
+        self.load_group(self.tr("Just added"))
         centralWidget.setLayout(vbox)
         self.setCentralWidget(centralWidget)
 
@@ -133,7 +132,7 @@ class PyTrailerWidget(QMainWindow):
         movieMenu = self.menuBar().addMenu(self.tr("&Movies"))
         i=1
         for cat, url in self.categories:
-            movieMenu.addAction(cat, self.slotCreate(cat),
+            movieMenu.addAction(cat, self.slot_create(cat),
                                 QKeySequence("F%d" % i))
             i = i + 1
         aboutMenu = self.menuBar().addMenu(self.tr("&Help"))
@@ -143,27 +142,69 @@ class PyTrailerWidget(QMainWindow):
                             QKeySequence(self.tr("Ctrl+A",
                                                  "Help|About PyQTrailer")))
 
+    def init_preloaders(self):
+        """Start all preloading processes and prepare data for them"""
+
+        # taskQueue - trailers to be cached go there
+        self.readAheadTaskQueue = multiprocessing.Queue()
+        # doneQueue - cached trailes come from there
+        self.readAheadDoneQueue = multiprocessing.Queue()
+
+        # trailers to be downloaded are put there
+        self.trailerDownloadQueue = multiprocessing.Queue()
+
+        # this is a url:DownloadStatus dictionary
+        self.trailerDownloadDict = multiprocessing.Manager().dict()
+
+        self.readAheadProcess = []
+        readAhead = int(self.config.get("DEFAULT","readAhead"))
+        for i in range(readAhead):
+            p = multiprocessing.Process(target=PyTrailerWidget.movie_readahead,
+                        args=(self.readAheadTaskQueue,
+                              self.readAheadDoneQueue,
+                              self.movie_cache))
+            p.start()
+            self.readAheadProcess.append(p)
+
+        self.downloader = TrailerDownloader(self.trailerDownloadQueue,
+                               self.trailerDownloadDict,
+                               int(self.config.get("DEFAULT","parallelDownload")))
+
+    def slot_create(self, group):
+        """This meta-function is used to create anonymous slot for each
+        trailer group."""
+        def slot():
+            self.load_group(group)
+        return slot
+
     def settings(self):
         d = PyTrailerSettings(self.config)
         if d.exec_() == QDialog.Accepted:
-            self.saveConfig()
+            self.save_config()
 
     def about(self):
         w = PyTrailerAbout(self)
         w.exec_()
 
-    def slotCreate(self, group):
-        def slot():
-            self.loadGroup(group)
-        return slot
 
-    def groupChange(self, button):
-        self.loadGroup(button.text())
+    def group_change(self, button):
+        self.load_group(button.text())
 
-    def unloadCurrentGroup(self):
+    def unload_current_group(self):
+        """Current trailer groups gets unloaded and preloader
+        processes stopped
+        """
         while not self.readAheadTaskQueue.empty():
             self.readAheadTaskQueue.get()
 
+        if self.list_loader:
+            self.list_loader.close()
+            self.list_loader = None
+
+        if self.list_loader_p:
+            self.list_loader_p.terminate()
+
+        # remove trailer widgets from main area
         widget = self.mainArea.takeAt(0)
         while widget != None:
             widget = widget.widget()
@@ -171,7 +212,7 @@ class PyTrailerWidget(QMainWindow):
             widget.close()
             widget = self.mainArea.takeAt(0)
 
-    def loadGroup(self, groupName):
+    def load_group(self, groupName):
         url = None
         for cat, catURL in self.categories:
             if cat == groupName:
@@ -185,40 +226,54 @@ class PyTrailerWidget(QMainWindow):
             else:
                 return
 
-        self.unloadCurrentGroup()
+        self.unload_current_group()
+        # loadID is used to identify what group task belonged to
+        # we can use it to make sure we don't display trailers from
+        # different group after being cached
         self.loadID = random.random()
         self.loading.setVisible(True)
+
+        # run loading in separate process
         self.list_loader , child_conn = multiprocessing.Pipe()
-        self.list_loader_p = multiprocessing.Process(target=movieListLoader,
+        self.list_loader_p = multiprocessing.Process(target=PyTrailerWidget.movielist_loader,
                                     args=(child_conn,url))
         self.list_loader_p.start()
 
 
-    def saveConfig(self):
+    def save_config(self):
         with open(self.configPath, 'w') as configfile:
             self.config.write(configfile)
 
-    def closeEvent(self, closeEvent):
-        self.saveConfig()
+    def closeEvent(self, close_event):
         self.downloader.stop()
         for p in self.readAheadProcess:
             p.terminate()
+        self.list_loader_p.terminate()
+        self.save_config()
+        self.save_cache()
 
-    def refreshMovies(self):
+    def refresh_movies(self):
+        # if we are loading new group and movie list is ready, get it
         if self.list_loader and self.list_loader.poll():
             self.movieList = self.list_loader.recv()
             self.display_group()
             self.list_loader_p.join()
+        # let's refresh status of trailer downloads
+        self.refresh_download_status()
 
-        self.refreshDownloadStatus()
+        # now to refreshing trailer widgets when posters, descriptions
+        # and trailer links have been cached
         while not self.readAheadDoneQueue.empty():
             i, updatedMovie, loadID = self.readAheadDoneQueue.get_nowait()
             if self.loadID != loadID:
+                # this means the cached movie comes from previous
+                # group (we must have changed group recently)
                 continue
             oldMovie = self.movieList[i]
             oldMovie.poster = updatedMovie.poster
             oldMovie.trailerLinks = updatedMovie.trailerLinks
             oldMovie.description = updatedMovie.description
+            self.add_to_cache(oldMovie)
             if oldMovie.title in self.movieDict:
                 w = self.movieDict[oldMovie.title]
                 if w is not None:
@@ -239,9 +294,14 @@ class PyTrailerWidget(QMainWindow):
                 self.loading.setVisible(False)
 
     def display_group(self):
+        """Makes the hard work to prepare current group for being
+        displayed
+
+        adds tasks to readahead process and creates widgets"""
         self.list_loader = None
         for i in range(len(self.movieList)):
             self.readAheadTaskQueue.put((i, self.movieList[i], self.loadID))
+
         try:
             filters = json.loads(self.config.get("DEFAULT","filters"))
         except ValueError:
@@ -249,16 +309,18 @@ class PyTrailerWidget(QMainWindow):
             import pickle
             filters = pickle.loads(self.config.get("DEFAULT","filters"))
 
+        # we create widgets but hide them until they are cached
         for movie in self.movieList:
             w=MovieItemWidget(movie, filters, self.scrollArea)
             w.setVisible(False)
-            w.downloadClicked.connect(self.downloadTrailer)
-            w.viewClicked.connect(self.viewTrailer)
+            w.downloadClicked.connect(self.download_trailer)
+            w.viewClicked.connect(self.view_trailer)
             self.movieDict[movie.title] = w
             self.mainArea.addWidget(w)
 
-    def refreshDownloadStatus(self):
-
+    def refresh_download_status(self):
+        """Makes sure download status is up to date
+        """
         for i in list(self.trailerDownloadDict.keys()):
             item = self.trailerDownloadDict[i]
             trailerName = item.url.split('/')[-1]
@@ -286,12 +348,15 @@ class PyTrailerWidget(QMainWindow):
         if len(list(self.trailerDownloadDict.keys())) and not self.statusView.isVisible():
             self.statusView.setVisible(True)
 
-    def downloadTrailer(self, url):
+    def download_trailer(self, url):
+        """Adds appropriate tasks to download trailer and sets initial
+        DownloadStatus"""
         self.trailerDownloadQueue.put((str(url),
                                       self.config.get("DEFAULT","downloadDir")))
         self.trailerDownloadDict[str(url)] = DownloadStatus(str(url),
                            DownloadStatus.WAITING)
-    def viewTrailer(self, url):
+
+    def view_trailer(self, url):
         player = self.config.get("DEFAULT","player").split(' ')
         for i in range(len(player)):
             if player[i] == '%a':
@@ -301,24 +366,61 @@ class PyTrailerWidget(QMainWindow):
 
         subprocess.Popen(player)
 
+    def add_to_cache(self, movie):
+        """Adds movie to disk cache"""
+        latestUpdate = movie.get_latest_trailer_date()
+        self.movie_cache[movie.baseURL] = (latestUpdate,
+                                           movie.poster,
+                                           movie.trailerLinks,
+                                           movie.description)
 
-def movieReadAhead(taskQueue, doneQueue):
-    """Function to be run in separate process,
-    caching additional movie information
-    """
-    while True:
-        i, movie, loadID = taskQueue.get()
+    def load_cache(self):
+        """Loads trailer info cache from disk"""
         try:
-            movie.poster
-            movie.trailerLinks
-            movie.description
-            doneQueue.put((i, movie, loadID))
-        except:
-            raise
+            with open(self.cachePath,"rb") as f:
+                self.movie_cache = pickle.load(f)
+        except IOError as e:
+            if e.errno == errno.ENOENT:
+                self.movie_cache = {}
+                self.movie_cache['last_update'] = 0
+            else:
+                raise
 
-def movieListLoader(conn, url):
-    conn.send(amt.getMoviesFromJSON(url))
-    conn.close()
+    def save_cache(self):
+        """Saves trailer info cache to disk"""
+        with open(self.cachePath,"wb") as f:
+            pickle.dump(self.movie_cache, f)
+
+
+    @staticmethod
+    def movie_readahead(taskQueue, doneQueue, cache):
+        """Function to be run in separate process,
+        caching additional movie information
+        """
+        while True:
+            i, movie, loadID = taskQueue.get()
+            try:
+                latestUpdate = movie.get_latest_trailer_date()
+                if movie.baseURL in cache and cache[movie.baseURL][0] >= latestUpdate:
+                    cached_data = cache[movie.baseURL]
+                    movie.poster = cached_data[1]
+                    movie.trailerLinks = cached_data[2]
+                    movie.description = cached_data[3]
+                else:
+                    movie.poster
+                    movie.trailerLinks
+                    movie.description
+                doneQueue.put((i, movie, loadID))
+            except:
+                raise
+
+    @staticmethod
+    def movielist_loader(conn, url):
+        """Function to be run in a separate process. It will send
+        movie list through pipe. Use poll() to verify that data is ready.
+        """
+        conn.send(amt.getMoviesFromJSON(url))
+        conn.close()
 
 
 if __name__ == "__main__":
